import numpy as np
from pymoc.utils import make_array


class SO_ML(object):
  r"""
  Southern Ocean Mixed Layer Model

  Instances of this class represent 1D representations of the meridional buoyancy structure
  in the Southern Ocean mixed layer, based on the solution to an advective-diffusive equation.
  Surface fluxes can be specified via a prescribed  fixed flux at the surface and/or
  a restoring buoyancy. Boundary conditions at the northern and southern boundaries
  of the Southern Ocean are determined internally based on the overturning streamfunction
  and buoyancy profile in the basin.

  """
  def __init__(
      self,
      y=None,    # grid (input)
      Ks=0.,    # hor. diffusivity (input)
      h=50.,    # ML depth (input)
      L=4e6,    # zonal width (input)
      surflux=0.,    # prescribed surface buoyancy flux (in m^2/s^3; input)
    # Notice that the first and last gridpoints are BCs and should have no flux
      rest_mask=0.,    # mask for surface restoring (1 where restoring is applied 0 elsewhere)
      b_rest=0.,    # surface buoyancy towards which we are restoring 
      v_pist=1.5 / 86400.,    # Piston velocity for restoring in SL (input)
      bs=0.0,    # Surface buoyancy (input, output)
      Psi_s=None    # Overturning in the ML (output)
  ):
    r"""
    Parameters
    ----------

    y : ndarray
        Uniform meridional mixed layer grid. Units: m
    Ks : float
         Horizontal diffusivity in the mixed layer. Units: 
    h : float
        Mixed layer depth. Units: m
    L : float
        Zonal length of the mixed layer. Units: m
    surflux : float, ndarray, or function; optional
              Prescribed surface buoyancy flux. Units: m\ :sup:`2`/s\ :sup:`3`
    rest_mask : float, ndarray, or function; optional
                Surface restoring mask, prescribed as 1 where restoring is desired, and 0 elsewhere.
    b_rest : float, ndarray, or function; optional
             Prescribed surface buoyancy profile towards which mixed layer buoyancy is restored. Units: 
    v_pist : float; optional
             Prescribed piston velocity for buoyancy restoration. Units: m/s
    bs : float, ndarray, or function; optional
         Initial meridional surface buoyancy profile in the mixed layer. Units:
    Psi_s : ndarray; optional
         Initial overturning transport in the mixed layer. Units: Sv 

    """
    # initialize grid:
    if isinstance(y, np.ndarray):
      self.y = y
    else:
      raise TypeError('y needs to be numpy array providing (regular) grid')

    self.Ks = Ks
    self.h = h
    self.L = L
    self.surflux = make_array(surflux, self.y, 'surflux')
    self.rest_mask = make_array(rest_mask, self.y, 'rest_mask')
    self.b_rest = make_array(b_rest, self.y, 'b_rest')
    self.v_pist = v_pist
    self.Psi_s = Psi_s
    self.bs = make_array(bs, self.y, 'bs')

  # def solve_equi(self):
  #Solve for equilibrium solution given inputs
  # raise TypeError('This functionality is not yet implemented')

  def set_boundary_conditions(self, b_basin, Psi_b):
    r"""
    Set the surface boundary conditions at the southern boundary of the Southern
    Ocean mixed layer. If upwelling takes place at the boundary, set the buoyancy
    there to match the densest upwelled density class from the adjoining basin.
    Otherwisee, impose a no-flux boundary condition.

    Parameters
    ----------

    b_basin : ndarray
              The buoyancy profile in the adjoining basin. Units:
    Psi_b : ndarray
            The transport streamfunction in the Southern Ocean. Units: Sv

    """
    if self.Psi_s[1] > 0:
      # set buoyancy at southern boundary to buoyancy of densest upwelling water
      self.bs[0] = b_basin[np.argwhere(Psi_b > 0)[0][0]]
    else:
      # no-flux BC
      self.bs[0] = self.bs[1]

  def calc_advective_tendency(self, dy):
    r"""
    Compute the advective transport tendency in the Southern Ocean mixed layer:

    .. math::
        \begin{aligned}
        \left[\partial_tb_{SO}\right]_{adv} &= -v_{SO}^\dagger\partial_yb_{SO} \\
        v_{SO}^\dagger &= \frac{\Psi_{SO}}{h\cdot L_x}
        \end{aligned}

    Parameters
    ----------

    dy: float
        The horizontal grid spacing for the mixed layer model.

    Returns
    -------

    dbdb_ad: ndarray
             An array representing the advecting tendency at each grid point.

    """

    # Notice that the current implementation assumes an evenly spaced grid!
    dbdt_ad = 0. * self.y
    indneg = self.Psi_s[1:-1] < 0.
    indpos = self.Psi_s[1:-1] > 0.
    dbdt_ad[1:-1][indneg] = -self.Psi_s[1:-1][indneg] * 1e6 * (
        self.bs[2:][indneg] - self.bs[1:-1][indneg]
    ) / self.h / self.L / dy
    dbdt_ad[1:-1][indpos] = -self.Psi_s[1:-1][indpos] * 1e6 * (
        self.bs[1:-1][indpos] - self.bs[:-2][indpos]
    ) / self.h / self.L / dy
    return dbdt_ad

  def calc_diffusion_matrix(self, s):
    r"""
    Compute the forward time centered space (FTCS) coeffient matrix for the implicit diffusion scheme, based on
    the stability criterion:

    .. math::
        s = K_s\cdot\frac{dt}{dy^2}

    Paramaers
    ---------
    s: float
       Stability criterion for FTCS implicit solution.

    Returns
    -------
    U: ndarray
       Coefficient matrix representing the system of linear equations being solved by the implicit diffusion scheme.

    """
    U = (
        np.diag(-s / 2. * np.ones(len(self.y) - 1), -1) +
        np.diag((1+s) * np.ones(len(self.y)), 0) +
        np.diag(-s / 2. * np.ones(len(self.y) - 1), 1)
    )
    U[0, 0] = 1
    U[0, 1] = 0
    U[-1, -2] = 0
    U[-1, -1] = 1

    return U

  def calc_implicit_diffusion(self, dy, dt):
    r"""
    Compute the diffusive transport in Southern Ocean mixed layer:

    .. math::
        \left[\partial_tb_{SO}\right]_{di\hspace{-0.1em}f\hspace{-0.2em}f} = \partial_y\left(K_s\partial_yb_{SO}\right)

    via an implicit scheme and apply it to the surface buoyancy.

    Parameters
    ----------
    
    dy: float
        The horizontal grid spacing for the mixed layer model.
    dt: float
        The timestep duration for the mixed layer model.

    Returns
    -------

    bs: ndarray
        Updated surface buoyancy, with the implicit diffusion applied.
    """

    s = self.Ks * dt / dy**2
    U = self.calc_diffusion_matrix(s)
    V = self.calc_diffusion_matrix(-s)
    Uinv = np.linalg.inv(U)

    return np.dot(np.dot(Uinv, V), self.bs)

  def advdiff(self, b_basin, Psi_b, dt):
    r"""
    Compute and apply advective-diffusive transport to the Souther Ocean Mixed Layer model,
    based on conditions in the adjoining basin.

    Parameters
    ----------

    b_basin: ndarray
              The buoyancy profile in the adjoining basin. Units:
    Psi_b: ndarray
            The transport streamfunction in the Southern Ocean. Units: Sv
    dt: float
        The timestep duration for the mixed layer model.

    """
    # update surface buoyancy profile via advect. and diff

    # First we need to determine Psi at the surface in the ML:
    # The first line here is a hack to reduce problems with interpolation
    # due to finite SO resolution when PsiSO goes to zero at the bottom
    #due to non-outcropping isopycnals
    # If Psi_b is zero at non-outcropping isopycnals, it can end up
    # very near zero also at the last (non-boundary) gridpoint
    # at the surface as a result of the interpolation procedure
    # - this is unphysical since psi by definition only vanishes on isopycnals
    # that don't outcrop (the problem is that this vanishing here is a step
    # function which messes with the interpolation.)
    # For the purpose of the interpolation to the surface we therefore set psi
    # on non-outcropping isopycnals to the last non-zero value above
    Psi_mod = Psi_b.copy()
    ind = np.nonzero(Psi_mod)[0][0]
    Psi_mod[:ind] = Psi_mod[ind]

    self.Psi_s = np.interp(self.bs, b_basin, Psi_mod)

    # The following makes sure that the circulation vanishes at latitudes
    # south of the densest surface buoyancy in case of non-monotonicity
    # of bs around Antarctica (a lense of lighter water around Antarctica
    # cannot connect to the channel and thus should have zero overturning.
    # Notice, however, that the model is not generally designed to properly
    # handle non-monotonic bs, so any such solution should treated with care)
    self.Psi_s[:np.argmin(self.bs)] = 0.
    self.Psi_s[
        0
    ] = 0.    # This value doesn't actually enter/matter, but zero overturning
    # at southern boundary makes more sense for diag purposes

    #set boundary conditions:
    self.set_boundary_conditions(b_basin, Psi_b)

    # Compute tendency due to surface b-flux
    dbdt_flux = self.surflux / self.h + self.rest_mask * self.v_pist / self.h * (
        self.b_rest - self.bs
    )

    # Compute advective tendency via upwind advection
    dy = self.y[1] - self.y[0]
    dbdt_ad = self.calc_advective_tendency(dy)

    # add tendencies from surface flux and advection
    self.bs = self.bs + dt * (dbdt_flux+dbdt_ad)

    #re-set southern boundary condition
    #(the above does not modify the boundary gridpoints,
    # but if we want to simulate no-flux bbc we need to adjust the boundary point accordingly):
    if self.Psi_s[1] <= 0:
      # no-flux BC
      self.bs[0] = self.bs[1]

    # Do implicit diffusion:
    self.bs = self.calc_implicit_diffusion(dy, dt)

    #re-set southern boundary condition:
    # this final re-set is just to pass back a state where boundary point is consistent with bcs
    # (preferable e.g. for computation of streamfunction)
    self.set_boundary_conditions(b_basin, Psi_b)

  def timestep(self, b_basin=None, Psi_b=None, dt=1.):
    r"""
    Integrate the mixed layer buoyancy profile for one timestep.

    Parameters
    ----------

    b_basin: ndarray
              The buoyancy profile in the adjoining basin. Units:
    Psi_b: ndarray
            The transport streamfunction in the Southern Ocean. Units: Sv
    dt: float
        The timestep duration for the mixed layer model.

    """

    #Integrate buoyancy profile evolution for one time-step
    if not isinstance(b_basin, np.ndarray):
      raise TypeError(
          'b_basin needs to be numpy array providing buoyancy levels in basin'
      )

    if not isinstance(Psi_b, np.ndarray):
      raise TypeError(
          'Psi_b needs to be numpy array providing overturning at buoyancy levels given by b_basin'
      )

    self.advdiff(b_basin=b_basin, Psi_b=Psi_b, dt=dt)



import numpy as np
from scipy import integrate, optimize
from pymoc.utils import make_func


class Psi_SO(object):
  r"""
  Southern Ocean Overturning Transport Model

  Instances of this class represent a 1D model of the overturning transport in
  the Southern Ocean interior, calculated based on a density profile in the
  adjoining basin, and local surface buoyancy and surface wind stress in the SO.

  """
  def __init__(
      self,
      z=None,    # vertical grid (array, in)
      y=None,    # horizontal grid (array, in)  
      b=None,    # buoyancy profile at northern end of ACC (function, array or float, in)
      bs=None,    # surface buoyancy (function, array or float, in)
      tau=None,    # surface wind stress (function, array or float, in)  
      f=1.2e-4,    # Coriolis parameter (in)
      rho=1030,    # Density of sea water (in)
      L=1e7,    # Zonal length of the ACC (in)
      KGM=1e3,    # GM coefficient (in)
      c=None,    # phase speed for F2010 BVP smoother of GM streamfunction
      bvp_with_Ek=False,    # if true, apply boundary condition that Psi_GM=-Psi_EK in F2010 BVP smoother 
      Hsill=None,    # height (in m above ocean floor) of the "sill", where Psi_Ek is tapered
      HEk=None,    # depth of surface Ekman layer
      Htapertop=None,    # A quadratic tapering of the GM streamfunction at the surface
      Htaperbot=None,    # A quadratic tapering of the GM streamfunction at the bottom
      smax=0.01,    # maximum slope for clipping of GM streamfunction
  ):
    r"""
    Parameters
    ----------

    z : ndarray
        Vertical depth levels of overturning grid. Units: m
    y : ndarray
        Meridional overturning grid. Units: m
    b : float, function, or ndarray
        Vertical buoyancy profile from the adjoining basin, on the north
        side of the ACC. Units: m/s\ :sup:`2`
    bs : float, function, or ndarray
         Surface level buoyancy boundary condition. Can be a constant,
         or an array or function in y. Units: m/s\ :sup:`2`
    tau : float, function, or ndarray
          Surface wind stress. Can be a constant, or an array or function
          in y. Units: N/m\ :sup:`2`
    f : float
        Coriolis parameter. Units s\ :sup:`-1`
    rho : float
          Density of sea water for Boussinesq approximation. Units: kg/m\ :sup:`3`
    L : float
        Zonal length of the modeled ACC. Units: m
    KGM : float
          Gent & McWilliams (GM) eddy diffusivity coefficient. Units: 
    c : float
        Phase speed cutoff for smoothing when solving the GM boundary value problem. Units: m/s 
    bvp_with_Ek : logical
            Whether to enforce the boundary condition that Psi_GM=-Psi_Ek at the ocean
            surface and bottom when solving the boundary value problem for the GM streamfunction.
    Hsill : float
            Height above the bottom at which the Ekman streamfunction is tapered. Units: m
    Hek : float
          Depth of the surface Ekman layer. Units: m
    Htapertop : float
                Height of the quadratic surface tapering layer for the GM streamfunction. Units: m
    Htaperbot : float
                Height of the quadratic bottom tapering layer for the GM streamfunction. Units: m
    smax : float
           Maximum slope of the GM streamfunction, above which Psi_GM is clipped. Units: m\ :sup:`-1`
    """

    # initialize grid:
    if isinstance(z, np.ndarray):
      self.z = z
    else:
      raise TypeError('z needs to be numpy array providing grid levels')

    if isinstance(y, np.ndarray):
      self.y = y
    else:
      raise TypeError(
          'y needs to be numpy array providing horizontal grid (or boundaries) of ACC'
      )

    self.b = make_func(b, self.z, 'b')
    self.bs = make_func(bs, self.y, 'bs')
    self.tau = make_func(tau, self.y, 'tau')
    self.f = f
    self.rho = rho
    self.L = L
    self.KGM = KGM
    self.c = c
    self.bvp_with_Ek = bvp_with_Ek
    self.Hsill = Hsill
    self.HEk = HEk
    self.Htapertop = Htapertop
    self.Htaperbot = Htaperbot
    self.smax = smax

  def ys(self, b):
    r"""
    Inversion function of :math:`bs\left(y\right)`. This is equivalent to the outcopping
    latitude of the isopycnal of density class :math:`b`.

    Parameters
    ----------

    b: float
       The surface buoyancy value for whose meridional location is being calculated.

    Returns
    -------

    ys : float
        The meridional location at which the surface buoyancy bs is equal to the supplied
        buoyancy value b.

    """
    def func(y):
      return self.bs(y) - b

    if b < np.min(self.bs(self.y)):
      # if b is smaller minimum bs, isopycnals don't outcrop and get handled separately
      return self.y[0] - 1e3
    if b > self.bs(self.y[-1]):
      # if b is larger than bs at northern end, return northernmost point:
      return self.y[-1]
    else:
      # if b in range of bs return ys(b):
      # Notice that this inversion is well defined only if bs is monotonically
      # increasing (past minind). Should probably add a check to make sure
      # this is the case...
      minind = np.argmin(self.bs(self.y))
      return optimize.brentq(func, self.y[minind], self.y[-1])

  def calc_N2(self):
    r"""
    Calculate the buouyancy (Brunt-Väisällä) frequency profile for the Southern Ocean

    Returns
    -------

    N2 : function
         A depth dependent function that returns the buoyancy frequency :math:`N^2` for a given depth :math:`z`.

    """

    dz = self.z[1:] - self.z[:-1]
    N2 = np.zeros(np.size(self.z))
    b = self.b(self.z)

    N2[1:-1] = (b[2:] - b[:-2]) / (dz[1:] + dz[:-1])
    N2[0] = (b[1] - b[0]) / dz[0]
    N2[-1] = (b[-1] - b[-2]) / dz[-1]

    return make_func(N2, self.z, 'N2')

  def calc_bottom_taper(self, H, z):
    r"""
    Calculate the quadratic tapering profile relative to the ocean floor.

    Parameters
    ----------

    H : float
        Height above the bottom at which the streamfunction is tapered. Units: m
    z : ndarray
        Vertical depth levels of overturning grid. Units: m

    Returns
    -------

    bottom_taper : ndarray
                   An array containing weights (0.0-1.0) corresponding to how much of the 
                   streamfunction should remain after tapering.

    """

    if H is not None:
      return 1. - np.maximum(z[0] + H - z, 0.)**2. / H**2.
    return 1.

  def calc_top_taper(self, H, z, scalar=True):
    r"""
    Calculate the quadratic tapering profile relative to the ocean surface.

    Parameters
    ----------

    H : float
        Depth from the surface at which the streamfunction is tapered. Units: m
    z : ndarray
        Vertical depth levels of overturning grid. Units: m

    Returns
    -------

    top_taper : ndarray
                An array containing weights (0.0-1.0) corresponding to how much of the 
                streamfunction should remain after tapering.

    """
    if H is not None:
      return 1 - np.maximum(z + H, 0)**2. / H**2.
    elif scalar:
      return 1.
    else:
      taper = np.ones(np.size(z))
      taper[-1] = 0.
      return taper

  def calc_Ekman(self):
    r"""
    Compute the Ekman transport from the wind stress averaged from the
    northern boundary of the domain to the latitude of the northernmost
    outcropped isopycnal.

    .. math::
      \Psi_{Ek} = -\frac{\tau L_x}{\rho_of_{SO}}

    Returns
    -------

    Psi_Ek : ndarray
             An array representing the meridional average of the Ekman transport at
             each vertical level of the Southern Ocean model.

    """

    tau_ave = 0 * self.z
    for ii in range(0, np.size(self.z)):
      y0 = self.ys(self.b(self.z[ii]))    # outcrop latitude
      tau_ave[ii] = np.mean(self.tau(np.linspace(y0, self.y[-1], 100)))

    silltaper = self.calc_bottom_taper(self.Hsill, self.z)
    Ektaper = self.calc_top_taper(self.HEk, self.z, scalar=False)
    return tau_ave / self.f / self.rho * self.L * silltaper * Ektaper

  def bc_GM(self, ya, yb):
    r"""
    Calculate the residuals of boundary conditions for the eddy driven transport
    boundary value problem.

    Parameters
    ----------

    ya : ndarray
         Bottom boundary condition. Units: Sv
    yb : ndarray
         Surface boundary condition. Units: Sv

    Returns
    -------

    bc : ndarray
         If bvp_with_Ek is false, an array containing the bottom boundary condition
         ya[0] and surface boundary condition yb[0].
         If bvp_with_Ek is true, an array containing the residuals of the supplied
         boundary conditions and the value of the Ekman transport :math:`\Psi_{Ek}`
         at those boundaries.
         
    """

    if self.bvp_with_Ek:
      return np.array([
          ya[0] + self.Psi_Ek[0] * 1e6, yb[0] + self.Psi_Ek[-1] * 1e6
      ])
    else:
      return np.array([ya[0], yb[0]])

  def calc_GM(self):
    r"""
    Compute the eddy (Gent & Mcwilliams) transport based on the meridionally
    averaged isopycnal slope.

    .. math::
      \begin{aligned}
      \Psi_{GM} &= K_{GM}\cdot s \\
      s\left(b\right) &\equiv \frac{z_B\left(b\right)}{L_y - y_{SO}\left(b\right)}
      \end{aligned}

    Where :math:`z_B\left(b\right)` is the depth of isopycnals of density class :math:`b`
    in the adjoining basin, and :math:`y_{SO}\left(b\right)` is the outcropping latitude
    of isopycnals of density class :math:`b` in the Southern Ocean, available via
    :meth:`pymoc.modules.Psi_SO.ys`.

    Returns
    -------

    Psi_GM : ndarray
             An array representing the meridional average of the eddy transport at
             each vertical level of the Southern Ocean model.

    """

    dy_atz = 0 * self.z
    eps = 0.1    # minimum dy (in meters) (to avoid div. by 0)
    for ii in range(0, np.size(self.z)):
      dy_atz[ii] = max(self.y[-1] - self.ys(self.b(self.z[ii])), eps)
    bottaper = self.calc_bottom_taper(self.Htaperbot, self.z)
    toptaper = self.calc_top_taper(self.Htapertop, self.z)
    if self.c is not None:
      temp = make_func(
          self.KGM * self.z / dy_atz * self.L * toptaper * bottaper, self.z,
          'psiGM'
      )
      N2 = self.calc_N2()

      def ode(z, y):
        return np.vstack((y[1], N2(z) / self.c**2. * (y[0] - temp(z))))

      #Solve the boundary value problem
      res = integrate.solve_bvp(
          ode, self.bc_GM, self.z, np.zeros((2, np.size(self.z)))
      )
      # return solution interpolated onto original grid
      temp = res.sol(self.z)[0, :]
    else:
      temp = self.KGM * np.maximum(
          self.z / dy_atz, -self.smax
      ) * self.L * toptaper * bottaper
    # limit Psi_GM to -Psi_Ek on isopycnals that don't outcrop:
    idx = dy_atz > self.y[-1] - self.y[0]
    temp[idx] = np.maximum(temp[idx], -self.Psi_Ek[idx] * 1e6)
    return temp

  def solve(self):
    r"""
    Compute the residual overturning transport in the Southern Ocean.

    .. math::
      \Psi_{SO} = \Psi_{Ek} + \Psi_{GM}

    Returns
    -------

    Psi : ndarray
          An array representing the meridional average of the residual overturning
          transport at each vertical level of the Southern Ocean. This is a scaled
          sum of the Ekman and eddy transports.
    """

    self.Psi_Ek = self.calc_Ekman() / 1e6
    self.Psi_GM = self.calc_GM() / 1e6
    self.Psi = self.Psi_Ek + self.Psi_GM
    # Notice that the bottom boundary is somewhat poorly defined, as it is
    # only for BC. To avoid random fluctuation in bottom Psi, we here simply
    # set it to value of last gridpoint above
    self.Psi[0] = self.Psi[1]

  def update(self, b=None, bs=None):
    r"""
    Update the vertical buoyancy profile and surface buoyancy, based on changes
    in the adjoining basin and/or in the surface boundary conditions.

    Parameters
    ----------

    b : float, function, or ndarray
        Vertical buoyancy profile from the adjoining basin, on the north
        side of the ACC. Units: m/s\ :sup:`2`
    bs : float, function, or ndarray
         Surface level buoyancy boundary condition. Can be a constant,
         or an array or function in y. Units: m/s\ :sup:`2`
    """

    if b is not None:
      self.b = make_func(b, self.z, 'b')
    if bs is not None:
      self.bs = make_func(bs, self.y, 'bs')



'''
This script defines a model class that can be used to solve a 1D column model
for the MOC.
 
The model is written in terms of a boundary value problem
 
In dimensional units, the ODE is
d_{zzzz}(\Psi_{N}) = (\kappa A)^{-1}(\Psi_{N} - \Psi_{SO} - A d_z(\kappa))d_{zzz}(\Psi_N)

This equation is solved subject to the boundary conditions:
(1) \Psi_N(0) = 0
(2) \Psi_N(-H) = 0 
(3) b(0)=-f \partial_{zz} \Psi_N (0) = b_s
(4) either  b(-H)=-f \partial_{zz} \Psi_N (H) = b_bot
    or d_z b(-H) = -f d_{zzz} \Psi_N (-H) =  (A \kappa(-H))^{-1} B_{int}
Where H is the total depth of the upper cell, which can also be solved for
with the additional BC that:
(5) d_z\psi_N(-H) = 0

The solution is found by non-dimensionalizing the equations using H and f as length and time scales
The model is then solved between z^*=z/H=0..1
Notice that H then appears as a parameter in the equations.
5 boundary conditions are needed if we also want to solve for the parameter H
'''

import numpy as np
from scipy import integrate
from pymoc.utils import check_numpy_version


class Equi_Column(object):
  def __init__(
      self,
      f=1.2e-4,    # Coriolis parameter (input)
      b_s=0.025,    # surface buoyancy (input)
      b_bot=None,    # bottom buoyancy (input)
      B_int=3e3,    # integrated downward buoyancy flux at the bottom of the upper cell (input)
      A=7e13,    # Area of the basin (input) 
      nz=100,    # minimum number of vert. layers for numerical solver (input)
      sol_init=None,    # Initial conditions for ODE solver (input)
      H_guess=1500.,    # guess for depth if solved for (input)
      kappa=6e-5,    # Diapycnal diffusivity (input; can be const., function, or array on grid given by z)
      dkappa_dz=None,    # Vertical derivative of diffusivity profile (input; function or nothing)
      psi_so=None,    # SO streamfunction (input; function or array on grid given by z)
      z=None,    # vertical grid for I/O (input / output) 
      psi=None,    # streamfunction (output)
      b=None,    # streamfunction (output)
      H=None,    # depth of cell (input / output)
  ):

    self.f = f
    self.A = A
    self.H = H
    self.H_guess = H_guess
    self.z = z
    self.zi = np.asarray(
        np.linspace(-1, 0, nz)
    )    # grid for initial conditions for solver

    if not callable(dkappa_dz) and (
        callable(kappa) or isinstance(kappa, np.ndarray)
    ) and not check_numpy_version():
      raise ImportError(
          'You need NumPy version 1.13.0 or later if you want to automatically compute dkappa_dz. Please upgrade your NumPy libary.'
      )

    self.kappa = self.init_kappa(kappa)
    self.dkappa_dz = self.init_dkappa_dz(kappa, dkappa_dz)
    self.init_psi_so(psi_so)
    self.init_b_boundaries(b_s, b_bot, B_int)
    self.sol_init = sol_init if sol_init is not None else self.calc_sol_init(
        sol_init, nz, b_bot
    )

  def init_kappa(self, kappa):
    # Initialize vertical diffusivity profile:
    if callable(kappa):
      return lambda z, H: kappa(z * H) / (
          H**2 * self.f
      )    # non-dimensionalize (incl. norm. of vertical coordinate)
    elif isinstance(kappa, np.ndarray):
      return lambda z, H: np.interp(z * H, self.z, kappa) / (H**2 * self.f)
    else:
      return lambda z, H: kappa / (H**2 * self.f)

  def init_dkappa_dz(self, kappa, dkappa_dz=None):
    if callable(kappa) and callable(dkappa_dz):
      return lambda z, H: dkappa_dz(z * H) / (H * self.f)
    elif callable(kappa):
      return lambda z, H: np.gradient(kappa(z * H), z * H) / (H * self.f)
    elif isinstance(kappa, np.ndarray):
      dkappa_dz = np.gradient(kappa, self.z)
      return lambda z, H: np.interp(z * H, self.z, dkappa_dz) / (H * self.f)
    else:
      return lambda z, H: 0

  def calc_sol_init(self, sol_init, nz=None, b_bot=None):
    # Set initial conditions for ODE solver
    if nz is None:
      nz = len(self.z)
    b_init = -100.0 if b_bot is not None else -self.bz(1500.)

    sol_init = np.zeros((4, nz))
    sol_init[0, :] = np.ones((nz))
    sol_init[2, :] = 0.0 * np.ones((nz))
    sol_init[3, :] = b_init * np.ones((nz))
    return sol_init

  # Initialize Southern Ocean Streamfunction
  def init_psi_so(self, psi_so=None):
    if callable(psi_so):
      self.psi_so = lambda z, H: psi_so(z * H) / (
          self.f * H**3
      )    # non-dimensionalize (incl. norm. of vertical coordinate)
    elif isinstance(psi_so, np.ndarray):
      self.psi_so = lambda z, H: np.interp(z * H, self.z, psi_so
                                           ) / (self.f * H**3)
    else:
      self.psi_so = lambda z, H: 0

  # initialize non-dimensional surface buoyancy and bottom buoyancy
  # or abyssal buoyancy flux boundary condition
  def init_b_boundaries(self, b_s, b_bot=None, B_int=None):
    if b_bot is None and B_int is None:
      raise Exception(
          'You need to specify either b_bot or B_int for bottom boundary condition'
      )
    self.bs = -b_s / self.f**2
    if b_bot is not None:
      self.b_bot = -b_bot / self.f**2
    else:
      self.B_int = B_int

  def alpha(self, z, H):
    #return factor on the RHS of ODE
    return H**2 / (self.A * self.kappa(z, H))

  def bz(self, H):
    # return the properly non-dimensionalized stratification at the bottom of the cell
    return self.B_int / (self.f**3 * H**2 * self.A * self.kappa(-1, H))

  def bc(self, ya, yb, p=None):
    try:
      bbot_set = getattr(self, 'b_bot', None) is not None
      y = np.array([ya[0], yb[0]])
      if self.H is None:
        y = np.append(y, ya[1])
      d = p[0] if self.H is None else self.H
      if bbot_set:
        y = np.append(y, ya[2] - self.b_bot / d)
      else:
        y = np.append(y, ya[3] + self.bz(d))
      y = np.append(y, yb[2] - self.bs / d)

      return y
    except TypeError:
      raise TypeError(
          'Must provide a p array if column does not have an H value'
      )

  def ode(self, z, y, p=None):
    #return the ODE to be solved
    if self.H is None and p is not None and len(p) > 0:
      H = p[0]
    elif self.H is not None:
      H = self.H
    else:
      raise TypeError(
          'Must provide a p array if column does not have an H value'
      )
    return np.vstack((
        y[1], y[2], y[3], self.alpha(z, H) * y[3] *
        (y[0] - self.psi_so(z, H) - self.A * self.dkappa_dz(z, H) / (H**2))
    ))

  def solve(self):
    #Solve the boundary value problem
    if self.H is None:
      res = integrate.solve_bvp(
          self.ode, self.bc, self.zi, self.sol_init, p=[self.H_guess]
      )
      self.H = res.p[0]
    else:
      res = integrate.solve_bvp(
          self.ode, self.bc, self.zi, self.sol_init, p=None
      )

    # if self.z does not yet exist use mesh from solver:
    if self.z is None:
      self.z = res.x * self.H
      self.psi = res.y[0, :] * self.f * self.H**3 / 1e6
      self.b = -res.y[2, :] * self.f**2 * self.H
    # if self.z does exist, compute solution at those points (setting all points below z=-H to NaN):
    else:
      self.psi = res.sol(self.z / self.H)[0, :] * self.f * self.H**3 / 1e6
      self.psi[self.z < -self.H] = np.NaN
      self.b = -res.sol(self.z / self.H)[2, :] * self.f**2 * self.H
      self.b[self.z < -self.H] = np.NaN



import numpy as np


def check_numpy_version():
  # check numpy version (version >= 1.13 needed to automatically compute db/dz)
  v = [int(i) for i in np.version.version.split('.')]
  if v[0] <= 1 and v[1] < 13:
    return False
  return True



'''
An class to interpolate the buoyancy field in the SO along lines
with constant slope - only to mke fancy looking plots
'''

import sys
import numpy as np
from scipy.optimize import brenth
from pymoc.utils import gridit, make_func


class Interpolate_channel(object):
  def __init__(
      self,
      y=None,    # y-grid
      z=None,    # z-grid
      bs=None,    # surface buoyancy
      bn=None,    # buoyancy profile in the north
  ):

    # initialize grid:
    if isinstance(y, np.ndarray):
      self.y = y
    else:
      raise TypeError('y needs to be numpy array providing grid levels')
    if isinstance(z, np.ndarray):
      self.z = z
    else:
      raise TypeError('z needs to be numpy array providing grid levels')

    self.bs = self.make_func(bs, 'bs', self.y)
    self.bn = self.make_func(bn, 'bn', self.z)

  def make_func(self, myst, name, xin):
    return make_func(myst, xin, name)

  def __call__(self, y, z):
    l = self.y[-1]
    if z == 0.0 and y == l:
      # slope ill defined at (l,0); evaluate infinitesimally below the surface:
      z = -1e-10

    def f2(x):
      # function to help determine slope at bottom of vent. region
      return self.bn(x) - self.bs(0)

    def f(x):
      # function to help determine slope in vent. region
      return self.bn(z - x * (l-y)) - self.bs(y + z/x)

    # first determine slope at bottom of vent. region here
    sbot = -brenth(f2, self.z[0], 0.) / l
    # than set slope for stuff above and below...
    if -z > sbot * y:
      s = sbot
    else:
      s = brenth(f, 1.e-12, 1.0)
    return self.bn(z - s * (l-y))

  def gridit(self):
    return gridit(self.y, self.z, self)



'''
This script defines a model class that can be used to compute the
overturning circulation between two columns, given buoyancy profiles
in the columns.
 
The model assumes a thermal-wind based equation for the overturning circulation
as in Nikurashin and Vallis (2012):
d_{zz}(\Psi) = f^{-1} (b_{2} - b_{1})

This equation is solved subject to the boundary conditions that
\Psi(0) = \Psi(-H) = 0 
(these BCs are different to NV2012)

An upwind isopycnal mapping is used to compute the isopycnal overturning transport
'''

import numpy as np
from scipy import integrate
from matplotlib import pyplot as plt
from pymoc.utils import make_func, make_array


class Psi_Thermwind(object):
  def __init__(
      self,
      f=1.2e-4,    # Coriolis parameter (input)
      z=None,    # grid (input)
      sol_init=None,    # Initial conditions for ODE solver (input)
      b1=None,    # Buoyancy in the basin (input, output)
      b2=0.,    # Buoyancy in the deep water formation region (input, output)
      Psi=None,    # Streamfunction (output) 
  ):

    self.f = f
    # initialize grid:
    if isinstance(z, np.ndarray):
      self.z = z
      nz = np.size(z)
    else:
      raise TypeError('z needs to be numpy array providing grid levels')

    self.b1 = self.make_func(b1, 'b1', self.z)
    self.b2 = self.make_func(b2, 'b2', self.z)

    # Set initial conditions for BVP solver
    if sol_init is None:
      self.sol_init = np.zeros((2, nz))
    else:
      self.sol_init = sol_init

  # end of init

  def make_func(self, myst, name, zin):
    return make_func(myst, zin, name)

  def make_array(self, myst, name):
    return make_array(myst, self.z, name)

  def bc(self, ya, yb):
    #return the boundary conditions
    return np.array([ya[0], yb[0]])

  def ode(self, z, y):
    #return the equation to be solved
    return np.vstack((y[1], 1. / self.f * (self.b2(z) - self.b1(z))))

  def solve(self):
    #Solve the boundary value problem
    # Note: The solution to this BVP is a relatively straightforward integral
    # it would probably be faster to just code it up that way.
    res = integrate.solve_bvp(self.ode, self.bc, self.z, self.sol_init)
    # interpolate solution for overturning circulation onto original grid (and change units to SV)
    self.Psi = res.sol(self.z)[0, :] / 1e6

  def Psib(self, nb=500):
    # map overturning into isopycnal space:
    b1 = self.make_array(self.b1, 'b1')
    b2 = self.make_array(self.b2, 'b2')
    bmin = min(np.min(b1), np.min(b2))
    bmax = max(np.max(b1), np.max(b2))
    self.bgrid = np.linspace(bmin, bmax, nb)
    udydz = -(self.Psi[1:] - self.Psi[:-1])
    psib = 0. * self.bgrid
    bup_bot = b1[:-1].copy()
    bup_top = b1[1:].copy()
    idx = udydz < 0
    bup_bot[idx] = b2[:-1][idx]
    bup_top[idx] = b2[1:][idx]
    for i in range(0, len(self.bgrid)):
      mask = np.clip((bup_top - self.bgrid[i]) / (bup_top-bup_bot), 0., 1.)
      psib[i] = np.sum(mask * udydz)
    return psib

  def Psibz(self, nb=500):
    # map isopycnal overturning back into isopycnal-depth space of each column
    psib = self.Psib(nb)
    # This does a linear interploation in b:
    return [
        np.interp(self.b1(self.z), self.bgrid, psib),
        np.interp(self.b2(self.z), self.bgrid, psib)
    ]
    # Ths instead first estimates the depth levels for the bgrid and then does linear interpolation in z
    # either has pros and cons depending on the situation...
    #z1_of_bgrid=np.interp(self.bgrid,self.b1(self.z),self.z)
    #z2_of_bgrid=np.interp(self.bgrid,self.b2(self.z),self.z)
    #return [np.interp(self.z,z1_of_bgrid,psib),np.interp(self.z,z2_of_bgrid,psib)]

  def update(self, b1=None, b2=None):
    # update buoyancy profiles
    if b1 is not None:
      self.b1 = self.make_func(b1, 'b1', self.z)
    if b2 is not None:
      self.b2 = self.make_func(b2, 'b2', self.z)



'''
A class to interpolate the buoyancy field in the North along lines
with constant slope - only to make fancy plots
'''

import sys
import numpy as np
from scipy.optimize import brenth
from pymoc.utils import gridit, make_func


class Interpolate_twocol(object):
  def __init__(
      self,
      y=None,    # y-grid
      z=None,    # z-grid
      bs=None,    # buoyancy profile in the south
      bn=None,    # buoyancy profile in the north
  ):

    # initialize grid:
    if isinstance(y, np.ndarray):
      self.y = y
    else:
      raise TypeError('y needs to be numpy array providing grid levels')
    if isinstance(z, np.ndarray):
      self.z = z
    else:
      raise TypeError('z needs to be numpy array providing grid levels')

    self.bs = self.make_func(bs, 'bs', self.z)
    self.bn = self.make_func(bn, 'bn', self.z)

  def make_func(self, myst, name, xin):
    return make_func(myst, xin, name)

  def __call__(self, y, z):
    l = self.y[-1]
    bsurf = self.make_func(
        self.y / l * self.bn(0) + (1 - self.y / l) * self.bs(0), 'bsurf',
        self.y
    )
    if z == 0 and y == 0:
      # slope ill defined at (0,0); evaluate infinitesimally below the surface:
      z = -0.01
    if z == self.z[0]:
      # slope also potentially ill defined at z= -H; evaluatejust above the bottom:
      z = 0.9999 * self.z[0]

    def fint(x):
      # function to help determine slope at bottom of vent. region
      return self.bn(0) - self.bs(-x * l)

    def fup(x):
      # function to help determine slope in vent. region
      return self.bs(z - x*y) - bsurf(y - z/x)

    def fdeep(x):
      # function to help determine slope below vent. region
      return self.bs(z - x*y) - self.bn(z + x * (l-y))

    # first determine slope at bottom of vent. region here
    sbot = brenth(fint, 0., 1.)
    # than set slope for stuff above and below...
    if z > -sbot * (l-y):
      s = brenth(fup, 1e-10, 1.0)
    else:
      s = brenth(fdeep, -1.0, 1.0)
    return self.bs(z - s*y)

  def gridit(self):
    return gridit(self.y, self.z, self)



import numpy as np


def make_array(myst, axis, name):
  # turn mysterious object into array(if needed)
  if isinstance(myst, np.ndarray):
    return myst
  elif callable(myst):
    return myst(axis)
  elif isinstance(myst, float):
    return myst + 0*axis
  else:
    raise TypeError(name, 'needs to be either function, numpy array, or float')



import numpy as np


def gridit(x1, x2, f):
  n1 = len(x1)
  n2 = len(x2)
  array = np.zeros((n1, n2))
  for i in range(0, n1):
    for j in range(0, n2):
      array[i, j] = f(x1[i], x2[j])
  return array



import numpy as np
from scipy import integrate
from pymoc.utils import make_func, make_array, check_numpy_version


class Column(object):
  r"""
  Vertical Advection-Diffusion Column Model

  Instances of this class represent 1D representations of buoyancy in a water
  column governed by vertical advection and diffusion. The velocity
  profile is required as an input. The script can either compute the equilibrium
  buoyancy profile for a given vertical velocity profile and boundary conditions
  or compute the tendency and perform a time-step of given length.
  BCs have to be fixed buoyancy at the top and either fixed b or db/dz at the bottom
  The time-stepping version can also handle horizontal advection
  into the column. This is, however, not (yet) implemented for the equilibrium solver
  """
  def __init__(
      self,
      z=None,    # grid (input)
      kappa=None,    # diffusivity profile (input)
      bs=0.025,    # surface buoyancy bound. cond (input)
      bbot=0.0,    # bottom buoyancy boundary condition (input)  
      bzbot=None,    # bottom strat. as alternative boundary condition (input) 
      b=0.0,    # Buoyancy profile (input, output)
      Area=None,    # Horizontal area (can be function of depth)
      N2min=1e-7    # Minimum strat. for conv adjustment
  ):
    r"""
    Parameters
    ----------

    z : ndarray
        Vertical depth levels of column grid. Units: m
    kappa : float, function, or ndarray
            Vertical diffusivity profile. Units: m\ :sup:`2`/s
    bs : float
         Surface level buoyancy boundary condition. Units: m/s\ :sup:`2`
    bbot : float; optional
           Bottom level buoyancy boundary condition. Units: m/s\ :sup:`2`
    bzbot : float; optional
            Bottom level buoyancy stratification. Can be used as an alternative to **bbot**. Units: s\ :sup:`-2`
    b : float, function, or ndarray
        Initial vertical buoyancy profile. Recalculated on model run. Units: m/s
    Area : float, function, or ndarray
           Horizontal area of basin. Units: m\ :sup:`2`
    N2min : float; optional
            Minimum stratification for convective adjustment. Units: s\ :sup:`-1`
    """

    # initialize grid:
    if isinstance(z, np.ndarray) and len(z) > 0:
      self.z = z
    else:
      raise TypeError('z needs to be numpy array providing grid levels')

    self.kappa = make_func(kappa, self.z, 'kappa')
    self.Area = make_func(Area, self.z, 'Area')

    self.bs = bs
    self.bbot = bbot
    self.bzbot = bzbot

    self.N2min = N2min

    self.b = make_array(b, self.z, 'b')

    if check_numpy_version():
      self.bz = np.gradient(self.b, z)
    else:
      self.bz = 0. * z    # notice that this is just for initialization of ode solver

  def Akappa(self, z):
    r"""
    Compute the area integrated diffusivity :math:`A\kappa`
    at depth(s) z.

    Parameters
    ----------

    z : float or ndarray
        Vertical depth level(s) at which to retrieve the integrated diffusivity.
    
    Returns
    -------
    AKappa : float or ndarray
             If z is a number, a number corresponding to the integrated diffusivity :math:`A\kappa` at that depth.
             If z is an ndarray, an ndarray where each entry corresponds to the integrated diffusivity :math:`A\kappa`
             at the z value with the same index.

    """

    return self.Area(z) * self.kappa(z)

  def dAkappa_dz(self, z):
    r"""
    Compute the area integrated diffusivity gradient :math:`\partial_z\left(A\kappa\right)`
    at depth(s) z.

    Parameters
    ----------

    z : float or ndarray
        Vertical depth level(s) at which to retrieve the integrated diffusivity gradient.

    Returns
    -------

    dAkappa_dz : float or ndarray
                 If z is a number, a number corresponding to the ther vertical gradients
                 in the integrated diffusivity :math:`\partial_zA\kappa` at that depth.
                 If z is an ndarray, an ndarray where each entry corresponds to the vertical gradient
                 in the integrated diffusivity :math:`\partial_zA\kappa` at the z value with the same index.

    """

    if not check_numpy_version():
      raise ImportError(
          'You need NumPy version 1.13.0 or later. Please upgrade your NumPy libary.'
      )
    return np.gradient(self.Akappa(z), z)

  def bc(self, ya, yb):
    r"""
    Calculate the residuals of boundary conditions for the advective-diffusive
    boundary value problem.

    Parameters
    ----------

    ya : ndarray
         Bottom boundary condition. Units: m/s\ :sup:`2`
    yb : ndarray
         Surface boundary condition. Units: m/s\ :sup:`2`

    Returns
    -------

    bc : ndarray
         If the bottom buoyancy stratification is defined as a boundary condition,
         an array containing the residuals of the imposed and calculated bottom buoyancy
         stratification and surface buoyancy.
         If the bottom buoyancy stratification is undefined as a boundary condition,
         an array containing the residuals of the imposed and calculated bottom buoyancy
         and surface buoyancy.

    """

    if self.bzbot is None:
      return np.array([ya[0] - self.bbot, yb[0] - self.bs])
    else:
      return np.array([ya[1] - self.bzbot, yb[0] - self.bs])

  def ode(self, z, y):
    r"""
    Generate the ordinary differential equation for the equilibrium buoyancy profile,
    to be solved as a boundary value problem:

    :math:`\partial_tb\left(z\right)=-w^\dagger\partial_zb+\partial_z\left(\kappa_{e\!f\!f}\partial_zb\right)`

    Parameters
    ----------

    z : ndarray
        Vertical depth levels of column grid on which to solve the ode. Units: m
    y : ndarray
        Initial values for buoyancy and buoyancy gradient profiles.

    Returns
    -------
    ode : ndarray
          A vertically oriented array, containing the system of linear equations:

          .. math::
            \begin{aligned}
            \partial_zy_1 &= y_2 \\
            \partial_zy_2 &= wA - \partial_zy_2\cdot\frac{\partial_zA\kappa}{A\kappa}
            \end{aligned}

    """

    return np.vstack(
        (y[1], (self.wA(z) - self.dAkappa_dz(z)) / self.Akappa(z) * y[1])
    )

  def solve_equi(self, wA):
    r"""
    Solve for the equilibrium buoyancy profile, given a specified vertical
    velocity profile, and pre-set surface and bottom boundary conditions.

    Parameters
    ----------

    wA : ndarray
         Area integrated velocity profile for the equilibrium solution. Units: m\ :sup:`3`/s

    """

    self.wA = make_func(wA, self.z, 'w')
    sol_init = np.zeros((2, np.size(self.z)))
    sol_init[0, :] = self.b
    sol_init[1, :] = self.bz
    res = integrate.solve_bvp(self.ode, self.bc, self.z, sol_init)
    # interpolate solution for b and db/dz onto original grid
    self.b = res.sol(self.z)[0, :]
    self.bz = res.sol(self.z)[1, :]

  def vertadvdiff(self, wA, dt):
    r"""
    Calculate and apply the forcing from advection and diffusion on the vertical buoyancy
    profile, for the timestepping solution. This function implements an upwind advection
    scheme.

    Parameters
    ----------

    wA : float or ndarray
         Area integrated velocity profile for the timestepping solution. Units: m\ :sup:`3`/s
    dt : int
         Numerical timestep over which solution are iterated. Units: s

    """

    wA = make_array(wA, self.z, 'wA')
    dz = self.z[1:] - self.z[:-1]

    # apply boundary conditions:
    self.b[-1] = self.bs
    self.b[0] = self.bbot if self.bzbot is None else self.b[
        1] - self.bzbot * dz[0]

    bz = (self.b[1:] - self.b[:-1]) / dz
    bz_up = bz[1:]
    bz_down = bz[:-1]
    bzz = (bz_up-bz_down) / (0.5 * (dz[1:] + dz[:-1]))

    #upwind advection:
    weff = wA - self.dAkappa_dz(self.z)
    bz = bz_down
    bz[weff[1:-1] < 0] = bz_up[weff[1:-1] < 0]

    db_dt = (
        -weff[1:-1] * bz / self.Area(self.z[1:-1]) +
        self.kappa(self.z[1:-1]) * bzz
    )
    self.b[1:-1] = self.b[1:-1] + dt*db_dt

  def convect(self):
    r"""
    Carry out downward convective adustment of the vertical buoyancy profile to
    the minimum stratification, :math:`N^2_{m\!i\!n}`. This adjustment assumes a fixed surface 
    buoyancy boundary condition.

    """
    # do convective adjustment to minimum strat N2min
    # notice that this parameterization currently only handles convection
    # from the top down which is the only case we really encounter here...
    # it also assumes a fixed surface buoyancy BC, and hence any
    # surface heat flux that's required to adjust the buoyancy of the
    # convecting column:
    ind = self.b > self.b[-1] + self.N2min * self.z
    self.b[ind] = self.b[-1] + self.N2min * self.z[ind]
    # Below is an energy conserving version that could be used
    # for model formulations without fixed surface b. But for fixed surface
    # b, the simpler version above is preferable as it deals better with long time steps
    # (for infinitesimal time-step and fixed surface b, the two are equivalent)
    # this version does currently also not include adjustment to finite strat.
    # dz=self.z[1:]-self.z[:-1];
    # dz=np.append(dz,dz[-1]);dz=np.insert(dz,0,dz[0])
    # dz=0.5*(dz[1:]+dz[0:-1]);
    # self.b[ind]=(np.mean(self.b[ind]*dz[ind]*self.Area(self.z[ind]))
    #            /np.mean(dz[ind]*self.Area(self.z[ind])) )

  def horadv(self, vdx_in, b_in, dt):
    r"""
    Carry out horizon buoyancy advection into the column model from an adjoining model,
    for the timestepping solution. This function implements an upwind advection scheme.

    Parameters
    ----------

    vdx_in : float or ndarray
             Total advective transport per unit height into the column for the timestepping
             solution. Positive values indicate transport into the column. Units: m\ :sup:`2`/s
    b_in : float or ndarray
           Buoyancy vales from the adjoining module for the timestepping solution. Units: m/s\ :sup:`2`
    dt : int
         Numerical timestep over which solution are iterated. Units: s

    """

    vdx_in = make_array(vdx_in, self.z, 'vdx_in')
    b_in = make_array(b_in, self.z, 'b_in')

    adv_idx = vdx_in > 0.0
    db = b_in - self.b

    self.b[adv_idx] = self.b[adv_idx] + dt * vdx_in[adv_idx] * db[
        adv_idx] / self.Area(self.z[adv_idx])

  def timestep(self, wA=0., dt=1., do_conv=False, vdx_in=None, b_in=None):
    r"""
    Carry out one timestep integration for the buoyancy profile, accounting
    for advective, diffusive, and convective effects.

    Parameters
    ----------

    wA : float or ndarray
         Area integrated velocity profile for the timestepping solution. Units: m\ :sup:`3`/s
    dt : int
         Numerical timestep over which solution are iterated. Units: s
    do_conv : logical
              Whether to carry out convective adjustment during model integration.
    vdx_in : float or ndarray
             Total advective transport per unit height into the column for the timestepping
             solution. Positive values indicate transport into the column. Units: m\ :sup:`2`/s
    b_in : float or ndarray
           Buoyancy vales from the adjoining module for the timestepping solution. Units: m/s\ :sup:`2`

    """

    self.vertadvdiff(wA=wA, dt=dt)
    if vdx_in is not None:
      # do horizontal advection: (optional)
      if b_in is not None:
        self.horadv(vdx_in=vdx_in, b_in=b_in, dt=dt)
      else:
        raise TypeError('b_in is needed if vdx_in is provided')
    if do_conv:
      # do convection: (optional)
      self.convect()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
  long_description = fh.read()

setup(
    name='py-moc',
    version='0.0.1rc4',
    description="A simple model suite for the MOC",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://pymoc.github.io',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



import numpy as np


def make_func(myst, axis, name):
  # turn mysterious object into callable function (if needed)
  if callable(myst):
    return myst
  elif isinstance(myst, np.ndarray):

    def funfun(x):
      return np.interp(x, axis, myst)

    return funfun
  elif isinstance(myst, float):

    def funfun(x):
      return myst + 0*x

    return funfun
  else:
    raise TypeError(name, 'needs to be either function, numpy array, or float')
